import typing

from weave.legacy.weave import api as weave

# TODO: Fix, these should be available from weave
from weave.legacy.weave import context_state, op_def, ops, panel, panels

loading_builtins_token = context_state.set_loading_built_ins()

# TODO: feels odd to call this "ops.Markdown".
# Maybe make it top level
# Or would types.Markdown by ok?
INTRO = ops.Markdown(
    """
# Welcome to Weave

*To understand intelligence: reflect!*

Weave is a new toolkit for constructing intelligent systems.

Click the links in the sidebar to get started.
"""
)

from weave.legacy.weave import registry_mem

op_org_name = registry_mem.memory_registry.get_op("user-name")


# TODO: using TypedDict here so we get automatic Object rendering
# But it'd be nice to use @weave.type() style so we have '.' attribute access instead
@weave.type()
class Ecosystem:
    _orgs: list[typing.Any]
    _packages: list[op_def.OpDef]
    _datasets: list[op_def.OpDef]
    _models: list[typing.Any]
    _ops: list[op_def.OpDef]

    @weave.op()
    def orgs(self) -> list[typing.Any]:
        return self._orgs

    @weave.op()
    def packages(self) -> list[op_def.OpDef]:
        return self._packages

    @weave.op()
    def datasets(self) -> list[op_def.OpDef]:
        return self._datasets

    @weave.op()
    def models(self) -> list[typing.Any]:
        return self._models

    @weave.op()
    def ops(self) -> list[op_def.OpDef]:
        return self._ops


# TODO: This should be entirely lazy, so that we don't actually need to
# load anything to construct it. We can load when the user browsers to specific
# objects.
@weave.op(name="op-ecosystem", render_info={"type": "function"})
def ecosystem() -> Ecosystem:
    from weave.legacy.weave import registry_mem

    return Ecosystem(
        _orgs=[],
        _packages=registry_mem.memory_registry.list_packages(),
        # This is not Weavey at all!  We should look for any registered ps of no arguments that produce DatasetCard type
        # We can do that query using... Weave, lazily. In fact, EcosystemType() doesn't need to store any data, just
        # render the panel. Which can lazily fetch / search for everything.
        # _datasets=[sklearn.ca_housing_dataset, torchvision.mnist],
        _datasets=[],
        _models=[],
        _ops=registry_mem.memory_registry.list_ops(),
    )


@weave.type()
class EcosystemPanel(panel.Panel):
    id = "EcosystemPanel"
    input_node: weave.Node[Ecosystem]

    @weave.op()
    def render(self) -> panels.Card:
        ecosystem = self.input_node
        return panels.Card(
            title="Weave",
            subtitle="",
            content=[
                panels.CardTab(
                    name="Intro",
                    content=panels.PanelMarkdown(INTRO),  # type: ignore
                ),
                panels.CardTab(
                    name="Organizations",
                    content=panels.Table(
                        weave.save(
                            [
                                "wandb",
                                "huggingface",
                                "deepmind",
                                "blueriver",
                                "google",
                                "openai",
                                "microsoft",
                            ],
                            name="org_list",
                        ),
                        columns=[
                            lambda org_name: panels.WeaveLink(
                                org_name,
                                lambda org_name: ops.entity(org_name),  # type: ignore
                            )  # type: ignore
                        ],
                    ),  # type: ignore
                ),
                panels.CardTab(
                    name="Packages",
                    content=panels.Table(
                        ecosystem.packages(),  # type: ignore
                        columns=[
                            lambda pack: panels.WeaveLink(
                                pack.op_name(),
                                vars={"package_root_op": ops.call_op(pack.op_name())},
                                to=lambda input, vars: vars["package_root_op"],
                            ),
                            lambda pack: pack.output_type(),
                        ],
                    ),
                ),
                panels.CardTab(
                    name="Datasets",
                    content=panels.Table(
                        ecosystem.datasets(),  # type: ignore
                        columns=[
                            lambda op: op.op_name(),
                            lambda op: op.output_type(),
                        ],
                    ),
                ),
                panels.CardTab(
                    name="Models",
                    content=panels.Table(ecosystem.models()),  # type: ignore
                ),
                panels.CardTab(
                    name="Ops",
                    content=panels.Table(
                        ecosystem.ops(),  # type: ignore
                        columns=[
                            lambda op: op.op_name(),
                            lambda op: op.output_type(),
                        ],
                    ),
                ),
            ],
        )


context_state.clear_loading_built_ins(loading_builtins_token)



import contextlib
import os
import typing

from weave.legacy.weave import weave_types as types

from . import file_base, file_util


class LocalFileType(file_base.FileBaseType):
    name = "local_file"

    def instance_to_dict(self, obj: "LocalFile") -> dict:
        return {
            "path": obj.path,
            "mtime": obj.mtime,
        }

    def instance_from_dict(self, d: dict) -> "LocalFile":
        return LocalFile(d["path"], d["mtime"])


class LocalFile(file_base.File):
    def __init__(
        self,
        path: str,
        mtime: typing.Optional[float] = None,
        extension: typing.Optional[str] = None,
    ) -> None:
        self.extension = extension
        if self.extension is None:
            self.extension = file_util.path_ext(path)
        self.path = path
        # Include mtime so that pure ops can consume us and hit cache
        # if the file has not been updated.
        self.mtime = mtime
        if self.mtime is None:
            self.mtime = os.path.getmtime(path)

    @contextlib.contextmanager
    def open(self, mode: str = "r") -> typing.Generator[typing.IO, None, None]:
        f = file_util.safe_open(self.path, mode)
        try:
            yield f
        finally:
            f.close()

    def size(self) -> int:
        return os.path.getsize(self.path)

    def get_local_path(self) -> str:
        return self.path

    def _file_contents_set(self, val: str) -> "LocalFile":
        with open(self.path, "w") as f:
            f.write(val)
        return self

    def _contents(self) -> str:
        with (open(self.path, encoding="ISO-8859-1")) as f:
            return f.read()


LocalFileType.instance_classes = LocalFile


class LocalDirType(types.ObjectType):
    _base_type = file_base.BaseDirType
    name = "localdir"

    def property_types(self) -> dict[str, types.Type]:
        return {
            "fullPath": types.String(),
            "size": types.Int(),
            "dirs": types.Dict(types.String(), file_base.SubDirType(LocalFileType())),
            "files": types.Dict(types.String(), LocalFileType()),
        }


class LocalDir(file_base.Dir):
    def path_info(
        self, path: str
    ) -> typing.Optional[typing.Union["LocalDir", LocalFile]]:
        return get_path_info(os.path.join(self.fullPath, path))


LocalDirType.instance_classes = LocalDir


def get_path_type(path: str) -> typing.Union[LocalDirType, LocalFileType]:
    file_util.check_path(path)
    if not os.path.exists(path):
        raise FileNotFoundError(f"Specified file or directory does not exist: '{path}'")
    elif os.path.isdir(path):
        return LocalDirType()
    else:
        ext = file_util.path_ext(path)
        return LocalFileType(extension=types.Const(types.String(), ext))


def get_path_info(path: str) -> typing.Optional[typing.Union["LocalDir", LocalFile]]:
    file_util.check_path(path)
    if not os.path.exists(path):
        return None
    elif os.path.isdir(path):
        sub_dirs = {}
        sub_files = {}
        for fname in os.listdir(path):
            full_path = os.path.join(path, fname)
            if os.path.isdir(full_path):
                subdir_dirs = {}
                subdir_files = {}
                for sub_fname in os.listdir(full_path):
                    sub_full_path = os.path.join(full_path, sub_fname)
                    if os.path.isdir(sub_full_path):
                        subdir_dirs[sub_fname] = 1
                    else:
                        subdir_files[sub_fname] = LocalFile(full_path)
                sub_dir = file_base.SubDir(
                    full_path, 10, subdir_dirs, subdir_files
                )  # TODO: size
                sub_dirs[fname] = sub_dir
            else:
                # sub_file = LocalFile(full_path, os.path.getsize(full_path), {}, {})
                sub_file = LocalFile(full_path)
                sub_files[fname] = sub_file
        return LocalDir(path, 111, sub_dirs, sub_files)
    else:
        return LocalFile(path)



import os
import string
import tempfile
import time
import typing
from pathlib import Path

from weave.legacy.weave import artifact_wandb as artifact_wandb
from weave.legacy.weave import environment
from weave.trace.refs import ObjectRef, parse_uri

try:
    from modal.cli.import_refs import import_stub
    from modal.config import config
    from modal.runner import deploy_stub
    from modal.serving import serve_stub
except ImportError:
    raise ImportError(
        "modal must be installed and configured: \n  pip install weave[modal]\n  modal setup"
    )


def compile(
    model_ref: str,
    project: str,
    reqs: list[str],
    auth_entity: typing.Optional[str] = None,
    secrets: typing.Optional[dict[str, str]] = None,
) -> Path:
    """Generates a modal py file and secret env vars to run the weave op"""
    dir = Path(tempfile.mkdtemp())
    with open(Path(__file__).parent / "stub.py", "r") as f:
        template = string.Template(f.read())
        src = template.substitute(
            {
                "REQUIREMENTS": '", "'.join(reqs),
                "MODEL_REF": model_ref,
                "PROJECT_NAME": project,
                "AUTH_ENTITY": auth_entity or "",
            }
        )

    with open(dir / "weave_model.py", "w") as f:
        f.write(src)
    with open(dir / ".env", "w") as f:
        if secrets is not None:
            for k, v in secrets.items():
                f.write(f"{k}={v}\n")
    return dir


def generate_modal_stub(
    model_ref: str,
    project_name: typing.Optional[str] = None,
    reqs: typing.Optional[list[str]] = None,
    auth_entity: typing.Optional[str] = None,
    secrets: typing.Optional[dict[str, str]] = None,
) -> str:
    """Generates a modal py file to run the weave op"""
    parsed_ref = parse_uri(model_ref)
    if project_name is None:
        if not isinstance(parsed_ref, ObjectRef):
            raise ValueError(f"Expected a weave object uri, got {type(parsed_ref)}")
        project_name = parsed_ref.project

    project = project_name or os.getenv("PROJECT_NAME")
    if project is None:
        raise ValueError(
            "project must be specified from command line or via the PROJECT_NAME env var"
        )
    reqs = reqs or []
    reqs.append("weave @ git+https://github.com/wandb/weave@weaveflow")
    reqs.append("fastapi>=0.104.0")
    return str(
        compile(model_ref, project, reqs, secrets=secrets, auth_entity=auth_entity)
        / "weave_model.py"
    )


def extract_secrets(model_ref: str) -> dict[str, str]:
    # TODO: get secrets from the weave op
    key = environment.weave_wandb_api_key()
    if key is None:
        secrets = {}
    else:
        secrets = {
            "WANDB_API_KEY": key,
        }
    openai_api_key = os.getenv("OPENAI_API_KEY")
    if openai_api_key:
        secrets["OPENAI_API_KEY"] = openai_api_key
    return secrets


def deploy(
    model_ref: str,
    wandb_project: typing.Optional[str] = None,
    auth_entity: typing.Optional[str] = None,
) -> None:
    """Deploy a model to the modal labs cloud."""
    ref = generate_modal_stub(
        model_ref,
        wandb_project,
        secrets=extract_secrets(model_ref),
        auth_entity=auth_entity,
    )
    stub = import_stub(ref)
    deploy_stub(stub, name=stub.name, environment_name=config.get("environment"))


def develop(model_ref: str, auth_entity: typing.Optional[str] = None) -> None:
    """Run a model for testing."""
    ref = generate_modal_stub(
        model_ref, secrets=extract_secrets(model_ref), auth_entity=auth_entity
    )
    print(f"Serving live code from: {ref}")
    stub = import_stub(ref)
    timeout = 1800
    with serve_stub(stub, ref, environment_name=config.get("environment")):
        while timeout > 0:
            t = min(timeout, 3600)
            time.sleep(t)
            timeout -= t



from __future__ import annotations

import threading
import typing
import warnings

import pyarrow as pa
from typing_extensions import (
    NotRequired,
    TypedDict,
)

import weave
from weave.legacy.weave import ops_arrow

UMAP_LIB = None

UMAP_LOCK = threading.Lock()


def get_umap():
    # Lazily load a cached version of UMAP - umap import
    # time is quite expensive so we want to do it once and
    # only when needed
    global UMAP_LIB
    if UMAP_LIB is None:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            import umap

            UMAP_LIB = umap
    return UMAP_LIB


class UMAPOptions(TypedDict):
    n_neighbors: NotRequired[int]
    min_dist: NotRequired[float]
    n_components: NotRequired[int]
    random_state: NotRequired[int]


@weave.op()
def umap_project(
    data: list[list[float]], options: typing.Optional[UMAPOptions]
) -> ops_arrow.ArrowWeaveList[list[float]]:
    if options == None:
        options = {}  # type: ignore
    model = get_umap().UMAP(**options)
    with UMAP_LOCK:
        umap_res = model.fit_transform(data)
    arrow_res = pa.array([a for a in umap_res])

    return ops_arrow.ArrowWeaveList(arrow_res)



import pytest

import weave
from weave.legacy.weave import ops_arrow


def test_dispatch_lambda():
    l = weave.save([1, 2, 3])
    filtered = l.filter(lambda x: x > 1)
    assert weave.use(filtered) == [2, 3]

    l = ops_arrow.to_weave_arrow([1, 2, 3])
    filtered = l.filter(lambda x: x > 1)
    assert weave.use(filtered).to_pylist_raw() == [2, 3]



import typing

import wandb

import weave
from weave.legacy.weave import context_state as _context
from weave.legacy.weave.wandb_interface.wandb_stream_table import StreamTable

from ...legacy.weave.panels_py import generator_templates

_loading_builtins_token = _context.set_loading_built_ins()


@weave.type()
class DummyBoardConfig:
    pass


board_input_type = weave.types.List(
    weave.types.TypedDict({"a": weave.types.optional(weave.types.String())})
)


@weave.op(  # type: ignore
    name="py_board-dummy_board",
    hidden=True,
    input_type={
        "input_node": weave.types.Function(
            {},
            board_input_type,
        )
    },
)
def dummy_board(
    input_node,
    config: typing.Optional[DummyBoardConfig] = None,
) -> weave.legacy.weave.panels.Group:
    assert board_input_type.assign_type(input_node.type)

    control_items = [
        weave.legacy.weave.panels.GroupPanel(
            input_node,
            id="data",
        ),
    ]

    return weave.legacy.weave.panels.Board(vars=control_items, panels=[])


generator_templates.template_registry.register(
    "py_board-dummy_board",
    "Simple Board",
    "Seed a board with a simple visualization of this table.",
)

_context.clear_loading_built_ins(_loading_builtins_token)


def assert_valid_node(node):
    _assert_valid_node_raw(node)

    # This part of the test simulates what happens when JS makes
    # a request since it is not guaranteed to know the correct type.
    node.type = weave.types.Any()
    _assert_valid_node_raw(node)


def _assert_valid_node_raw(node):
    templates_node = generator_templates.get_board_templates_for_node(node)

    templates = weave.use(templates_node)

    # Assert that the dummy board is successfully assigned
    assert len([t for t in templates if t["op_name"] == "py_board-dummy_board"]) == 1

    dummy_node = dummy_board(node, None)
    output_group = weave.use(dummy_node)

    # Assert the out template is successfully generated
    assert isinstance(output_group, weave.legacy.weave.panels.Group)

    data_node = output_group.config.items["sidebar"].config.items["data"]

    # assert that the node sent to the generator is the same as the node that is used
    # Note: this is a heuristic, but probably close enough
    assert isinstance(data_node, weave.legacy.weave.graph.OutputNode)
    assert data_node.from_op.name == node.from_op.name
    assert str(data_node) == str(node)


def assert_invalid_node(node):
    templates_node = generator_templates.get_board_templates_for_node(node)

    templates = weave.use(templates_node)

    assert len([t for t in templates if t["op_name"] == "py_board-dummy_board"]) == 0


def test_templates_for_run_logs_valid(user_by_api_key_in_env):
    run = wandb.init(project="project_exists")
    run.log({"a": "hello"})
    run.finish()

    run_history_node = (
        weave.legacy.weave.ops.project(run.entity, run.project).run(run.id).history()
    )

    assert_valid_node(run_history_node)


def test_templates_for_run_logs_invalid(user_by_api_key_in_env):
    run = wandb.init(project="project_exists")
    run.log({"a": 42})
    run.finish()

    run_history_node = (
        weave.legacy.weave.ops.project(run.entity, run.project).run(run.id).history()
    )

    assert_invalid_node(run_history_node)


def test_templates_for_logged_table_valid(user_by_api_key_in_env):
    run = wandb.init(project="project_exists")
    run.log({"table": wandb.Table(data=[["hello"]], columns=["a"])})
    run.finish()

    table_node = (
        weave.legacy.weave.ops.project(run.entity, run.project)
        .run(run.id)
        .summary()["table"]
        .table()
        .rows()
    )

    assert_valid_node(table_node)


def test_templates_for_logged_table_invalid(user_by_api_key_in_env):
    run = wandb.init(project="project_exists")
    run.log({"table": wandb.Table(data=[[42]], columns=["a"])})
    run.finish()

    table_node = (
        weave.legacy.weave.ops.project(run.entity, run.project)
        .run(run.id)
        .summary()["table"]
        .table()
        .rows()
    )

    assert_invalid_node(table_node)


def test_templates_for_stream_table_valid(user_by_api_key_in_env):
    st = StreamTable(
        "test_table",
        project_name="stream-tables",
        entity_name=user_by_api_key_in_env.username,
        _disable_async_file_stream=True,
    )
    st.log({"a": "hello"})
    st.finish()

    rows_node = st.rows()

    assert_valid_node(rows_node)


def test_templates_for_stream_table_invalid(user_by_api_key_in_env):
    st = StreamTable(
        "test_table",
        project_name="stream-tables",
        entity_name=user_by_api_key_in_env.username,
        _disable_async_file_stream=True,
    )
    st.log({"a": 1})
    st.finish()

    rows_node = st.rows()

    assert_invalid_node(rows_node)



import pytest

import weave


def test_cast_basic():
    valid_cast = weave.legacy.weave.ops.cast(1, weave.types.Int())
    assert valid_cast.type == weave.types.Int()
    assert weave.use(valid_cast) == 1

    invalid_cast = weave.legacy.weave.ops.cast(1, weave.types.String())
    assert invalid_cast.type == weave.types.String()
    with pytest.raises(weave.errors.WeaveTypeError):
        weave.use(invalid_cast)


def test_cast_json():
    s = weave.save('{"a": 1}')
    parsed = s.json_parse().cast(weave.types.TypedDict({"a": weave.types.Int()}))
    assert parsed.type == weave.types.TypedDict({"a": weave.types.Int()})
    assert weave.use(parsed) == {"a": 1}



import os
import random
import shutil
import tempfile

import numpy as np
import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient

import weave
from weave import context_state
from weave.trace import weave_init
from weave.trace.context import raise_on_captured_errors
from weave.trace_server import (
    clickhouse_trace_server_batched,
    sqlite_trace_server,
)
from weave.trace_server import trace_server_interface as tsi
from weave.trace_server_bindings import remote_http_trace_server

from .tests import fixture_fakewandb
from .tests.trace.trace_server_clickhouse_conftest import *
from .tests.wandb_system_tests_conftest import *
from .trace import autopatch

# Force testing to never report wandb sentry events
os.environ["WANDB_ERROR_REPORTING"] = "false"


class FakeTracer:
    def trace(*args, **kwargs):
        pass


def make_fake_tracer():
    return FakeTracer()


### End disable datadog engine tracing
### disable internet access


def guard(*args, **kwargs):
    raise Exception("I told you not to use the Internet!")


### End disable internet access

# Uncomment these two lines to disable internet access entirely.
# engine_trace.tracer = make_fake_tracer
# socket.socket = guard


@pytest.fixture()
def test_artifact_dir():
    return "/tmp/weave/pytest/%s" % os.environ.get("PYTEST_CURRENT_TEST")


def pytest_sessionfinish(session, exitstatus):
    if exitstatus == pytest.ExitCode.NO_TESTS_COLLECTED:
        print("No tests were selected. Exiting gracefully.")
        session.exitstatus = 0


def pytest_collection_modifyitems(config, items):
    # Add the weave_client marker to all tests that have a client fixture
    for item in items:
        if "client" in item.fixturenames:
            item.add_marker(pytest.mark.weave_client)

    # Get the job number from environment variable (0 for even tests, 1 for odd tests)
    job_num = config.getoption("--job-num", default=None)
    if job_num is None:
        return

    job_num = int(job_num)

    selected_items = []
    for index, item in enumerate(items):
        if index % 2 == job_num:
            selected_items.append(item)

    items[:] = selected_items


@pytest.fixture(autouse=True)
def always_raise_on_captured_errors():
    with raise_on_captured_errors():
        yield


@pytest.fixture(autouse=True)
def throw_on_error():
    os.environ["WEAVE_VALUE_OR_ERROR_DEBUG"] = "true"
    yield
    del os.environ["WEAVE_VALUE_OR_ERROR_DEBUG"]


@pytest.fixture()
def cache_mode_minimal():
    os.environ["WEAVE_NO_CACHE"] = "true"
    yield
    del os.environ["WEAVE_NO_CACHE"]


@pytest.fixture()
def cereal_csv():
    with tempfile.TemporaryDirectory() as d:
        cereal_path = os.path.join(d, "cereal.csv")
        shutil.copy("testdata/cereal.csv", cereal_path)
        yield cereal_path


@pytest.fixture()
def fake_wandb():
    setup_response = fixture_fakewandb.setup()
    yield setup_response
    fixture_fakewandb.teardown(setup_response)


@pytest.fixture()
def fixed_random_seed():
    random.seed(8675309)
    np.random.seed(8675309)
    yield
    random.seed(None)
    np.random.seed(None)


@pytest.fixture()
def app():
    from . import weave_server

    app = weave_server.make_app()
    app.config.update(
        {
            "TESTING": True,
        }
    )

    yield app


@pytest.fixture()
def enable_touch_on_read():
    os.environ["WEAVE_ENABLE_TOUCH_ON_READ"] = "1"
    yield
    del os.environ["WEAVE_ENABLE_TOUCH_ON_READ"]


@pytest.fixture()
def consistent_table_col_ids():
    from weave.legacy.weave.panels import table_state

    with table_state.use_consistent_col_ids():
        yield


@pytest.fixture()
def strict_op_saving():
    with context_state.strict_op_saving(True):
        yield


class TestOnlyFlushingWeaveClient(weave_client.WeaveClient):
    """
    A WeaveClient that automatically flushes after every method call.

    This subclass overrides the behavior of the standard WeaveClient to ensure
    that all write operations are immediately flushed to the underlying storage
    before any subsequent read operation is performed. This is particularly
    useful in testing scenarios where data is written and then immediately read
    back, as it eliminates potential race conditions or inconsistencies that
    might arise from delayed writes.

    The flush operation is applied to all public methods (those not starting with
    an underscore) except for the 'flush' method itself, to avoid infinite recursion.
    This aggressive flushing strategy may impact performance and should primarily
    be used in testing environments rather than production scenarios.

    Note: Due to this, the test suite essentially "blocks" on every operation. So
    if we are to test timing, this might not be the best choice. As an alternative,
    we could explicitly call flush() in every single test that reads data, before the
    read operation(s) are performed.
    """

    def __getattribute__(self, name):
        self_super = super()
        attr = self_super.__getattribute__(name)

        if callable(attr) and not name.startswith("_") and name != "flush":

            def wrapper(*args, **kwargs):
                res = attr(*args, **kwargs)
                self_super._flush()
                return res

            return wrapper

        return attr


def make_server_recorder(server: tsi.TraceServerInterface):  # type: ignore
    """A wrapper around a trace server that records all attribute access.

    This is extremely helpful for tests to assert that a certain series of
    attribute accesses happen (or don't happen), and in order. We will
    probably want to make this a bit more sophisticated in the future, but
    this is a pretty good start.

    For example, you can do something like the followng to assert that various
    read operations do not happen!

    ```pyth
    access_log = client.server.attribute_access_log
    assert "table_query" not in access_log
    assert "obj_read" not in access_log
    assert "file_content_read" not in access_log
    ```
    """

    class ServerRecorder(type(server)):  # type: ignore
        attribute_access_log: list[str]

        def __init__(self, server: tsi.TraceServerInterface):  # type: ignore
            self.server = server
            self.attribute_access_log = []

        def __getattribute__(self, name):
            self_server = super().__getattribute__("server")
            access_log = super().__getattribute__("attribute_access_log")
            if name == "server":
                return self_server
            if name == "attribute_access_log":
                return access_log
            attr = self_server.__getattribute__(name)
            if name != "attribute_access_log":
                access_log.append(name)
            return attr

    return ServerRecorder(server)


@pytest.fixture()
def client(request) -> Generator[weave_client.WeaveClient, None, None]:
    inited_client = None
    weave_server_flag = request.config.getoption("--weave-server")
    server: tsi.TraceServerInterface
    entity = "shawn"
    project = "test-project"
    if weave_server_flag == "sqlite":
        sqlite_server = sqlite_trace_server.SqliteTraceServer(
            "file::memory:?cache=shared"
        )
        sqlite_server.drop_tables()
        sqlite_server.setup_tables()
        server = TestOnlyUserInjectingExternalTraceServer(
            sqlite_server, DummyIdConverter(), entity
        )
    elif weave_server_flag == "clickhouse":
        ch_server = clickhouse_trace_server_batched.ClickHouseTraceServer.from_env()
        ch_server.ch_client.command("DROP DATABASE IF EXISTS db_management")
        ch_server.ch_client.command("DROP DATABASE IF EXISTS default")
        ch_server._run_migrations()
        server = TestOnlyUserInjectingExternalTraceServer(
            ch_server, DummyIdConverter(), entity
        )
    elif weave_server_flag.startswith("http"):
        remote_server = remote_http_trace_server.RemoteHTTPTraceServer(
            weave_server_flag
        )
        server = remote_server
    elif weave_server_flag == ("prod"):
        inited_client = weave_init.init_weave("dev_testing")

    if inited_client is None:
        client = TestOnlyFlushingWeaveClient(
            entity, project, make_server_recorder(server)
        )
        inited_client = weave_init.InitializedClient(client)
        autopatch.autopatch()
    try:
        yield inited_client.client
    finally:
        inited_client.reset()


@pytest.fixture
def network_proxy_client(client):
    """
    This fixture is used to test the `RemoteHTTPTraceServer` class. There is
    almost no logic in this class, other than a little batching, so we typically
    skip it for simplicity. However, we can use this fixture to test such logic.
    It initializes a mini FastAPI app that proxies requests from the
    `RemoteHTTPTraceServer` to the underlying `client.server` object.

    We probably will want to flesh this out more in the future, but this is a
    starting point.
    """
    app = FastAPI()

    records = []

    @app.post("/table/create")
    def table_create(req: tsi.TableCreateReq) -> tsi.TableCreateRes:
        records.append(
            (
                "table_create",
                req,
            )
        )
        return client.server.table_create(req)

    @app.post("/table/update")
    def table_update(req: tsi.TableUpdateReq) -> tsi.TableUpdateRes:
        records.append(
            (
                "table_update",
                req,
            )
        )
        return client.server.table_update(req)

    with TestClient(app) as c:

        def post(url, data=None, json=None, **kwargs):
            kwargs.pop("stream", None)
            return c.post(url, data=data, json=json, **kwargs)

        orig_post = weave.trace_server.requests.post
        weave.trace_server.requests.post = post

        remote_client = remote_http_trace_server.RemoteHTTPTraceServer(
            trace_server_url=""
        )
        yield (client, remote_client, records)

        weave.trace_server.requests.post = orig_post



from pydantic import BaseModel, ValidationError

from . import trace_server_interface as tsi
from .errors import InvalidRequest
from .orm import Column, Table
from .validation import validate_purge_req_multiple, validate_purge_req_one

TABLE_FEEDBACK = Table(
    "feedback",
    [
        Column("id", "string"),
        Column("project_id", "string"),
        Column("weave_ref", "string"),
        Column("wb_user_id", "string"),
        Column("creator", "string", nullable=True),
        Column("created_at", "datetime"),
        Column("feedback_type", "string"),
        Column("payload", "json", db_name="payload_dump"),
    ],
)


FEEDBACK_PAYLOAD_SCHEMAS: dict[str, type[BaseModel]] = {
    "wandb.reaction.1": tsi.FeedbackPayloadReactionReq,
    "wandb.note.1": tsi.FeedbackPayloadNoteReq,
}


def validate_feedback_create_req(req: tsi.FeedbackCreateReq) -> None:
    payload_schema = FEEDBACK_PAYLOAD_SCHEMAS.get(req.feedback_type)
    if payload_schema:
        try:
            payload_schema(**req.payload)
        except ValidationError as e:
            raise InvalidRequest(
                f"Invalid payload for feedback_type {req.feedback_type}: {e}"
            )


MESSAGE_INVALID_FEEDBACK_PURGE = (
    "Can only purge feedback by specifying one or more feedback ids"
)


def validate_feedback_purge_req(req: tsi.FeedbackPurgeReq) -> None:
    """For safety, we currently only allow purging by feedback id."""
    expr = req.query.expr_.model_dump()
    keys = list(expr.keys())
    if len(keys) != 1:
        raise InvalidRequest(MESSAGE_INVALID_FEEDBACK_PURGE)
    if keys[0] == "eq_":
        validate_purge_req_one(expr, MESSAGE_INVALID_FEEDBACK_PURGE)
    elif keys[0] == "or_":
        validate_purge_req_multiple(expr["or_"], MESSAGE_INVALID_FEEDBACK_PURGE)
    else:
        raise InvalidRequest(MESSAGE_INVALID_FEEDBACK_PURGE)



import os
import typing

import pytest

import weave
from weave.legacy.weave import api, environment, execute, ops, weave_internal
from weave.legacy.weave import weave_types as types

from . import test_wb

execute_test_count_op_run_count = 0

from weave.legacy.weave import context_state as _context_state

_loading_builtins_token = _context_state.set_loading_built_ins()


@api.op(input_type={"x": types.Any()}, output_type=types.Number(), hidden=True)
def execute_test_count_op(x):
    global execute_test_count_op_run_count
    execute_test_count_op_run_count += 1
    return len(x)


REFINE_CALLED = 0


@weave.op()
def _test_execute_refining_op_refine(x: int) -> weave.types.Type:
    global REFINE_CALLED
    REFINE_CALLED += 1
    return weave.types.Int()


@weave.op(refine_output_type=_test_execute_refining_op_refine)
def _test_execute_refining_op(x: int) -> typing.Any:
    return x + 1


_context_state.clear_loading_built_ins(_loading_builtins_token)


def test_local_file_pure_cached(cereal_csv):
    # local_path() is impure, but the operations thereafter are pure
    # this test confirms that pure ops that come after impure ops hit cache
    global execute_test_count_op_run_count
    execute_test_count_op_run_count = 0
    # We should only execute execute_test_count_op once.
    count1 = api.use(execute_test_count_op(ops.local_path(cereal_csv).readcsv()))
    count2 = api.use(execute_test_count_op(ops.local_path(cereal_csv).readcsv()))
    assert count1 == count2
    assert execute_test_count_op_run_count == 1


def test_execute_no_cache():
    nine = weave_internal.make_const_node(types.Number(), 9)
    res = execute.execute_nodes([nine + 3], no_cache=True)
    assert res.unwrap() == [12]


@pytest.fixture()
def weave_cache_mode_minimal():
    orig_cache_mode = environment.cache_mode

    def _cache_mode():
        return environment.CacheMode.MINIMAL

    environment.cache_mode = _cache_mode
    yield
    environment.cache_mode = orig_cache_mode


def test_execute_cache_mode_minimal_no_recursive_refinement(weave_cache_mode_minimal):
    called_once = _test_execute_refining_op(1)
    assert REFINE_CALLED == 1
    assert called_once.type == weave.types.Int()
    called_twice = _test_execute_refining_op(called_once)
    # with cache mode minimal, we do get recursive refine while construct
    # graphs by calling ops!
    assert REFINE_CALLED == 3
    assert called_twice.type == weave.types.Int()

    result = weave.use(called_twice)
    assert result == 3
    # But, when we execute with cache mode minimal, we only call each refine
    # once.
    assert REFINE_CALLED == 5


def test_we_dont_over_execute(fake_wandb):
    fake_wandb.fake_api.add_mock(test_wb.table_mock1)
    cell_node = (
        ops.project("stacey", "mendeleev")
        .runs()
        .limit(1)
        .summary()["table"]
        .table()
        .rows()
        .dropna()
        .concat()
        .createIndexCheckpointTag()[5]["score_Amphibia"]
    )
    with execute.top_level_stats() as stats:
        assert weave.use(cell_node.indexCheckpoint()) == 5

    # If this check fails, please be very careful! It means that something
    # has changed in the engine that is causing us to over-execute nodes
    # and is an indication of a serious performance regression.
    summary = stats.summary()
    assert summary["count"] - summary["already_executed"] == 16


def table_mock_respecting_run_name(q, ndx):
    # this is a more realistic gql responder that will only return run displayName
    # if its selected.
    if q["gql"].definitions[0].name.value == "WeavePythonCG":
        # is project.runs.edges.node.displayName selected?
        project_field = q["gql"].definitions[0].selection_set.selections[0]
        assert project_field.name.value == "project"
        runs_field = project_field.selection_set.selections[2]
        assert runs_field.name.value == "runs"
        runs_node = runs_field.selection_set.selections[0].selection_set.selections[0]
        node_fields = runs_node.selection_set.selections
        display_name_selected = any(f.name.value == "displayName" for f in node_fields)

        if display_name_selected:
            return test_wb.workspace_response()
        else:
            return test_wb.workspace_response_no_run_displayname
    else:
        return test_wb.artifact_version_sdk_response


def test_outer_tags_propagate_on_cache_hit(fake_wandb):
    fake_wandb.fake_api.add_mock(table_mock_respecting_run_name)
    row = (
        ops.project("stacey", "mendeleev")
        .runs()
        .limit(50)
        .summary()
        .pick("table")
        .offset(0)
        .limit(2)[0]
        .table()
        .rows()
        .createIndexCheckpointTag()[0]
    )
    # The first query does not select run.displayName, but it caches
    # results along the way.
    weave.use(row)

    # The cached results include tags that have gql results (which don't
    # include displayName since it wasn't selected in the prior query).

    # Now select displayName in a new query. This query will hit cache.
    # But we should get the correct result because we propagate outer
    # tags even on cache hits.
    run_names_res = weave.use(row.run().name())
    assert run_names_res == "amber-glade-100"


@weave.op()
def expensive_op(x: int) -> int:
    return x + 10000


@pytest.mark.skip(reason="Disabled in favor of parallelism for the moment.")
def test_cache_column():
    os.environ["WEAVE_CACHE_MODE"] = "minimal"

    input_vals = list(range(10))
    expected_result = [{"x": x, "y": x + 10000} for x in input_vals]

    l = weave.save(input_vals)
    mapped = l.map(lambda x: weave.legacy.weave.ops.dict_(x=x, y=expensive_op(x)))
    res = weave.use(mapped)
    assert res == expected_result

    latest_obj = weave.use(
        weave.legacy.weave.ops.get("local-artifact:///run-op-expensive_op:latest/obj")
    )
    assert len(latest_obj) == len(input_vals)
    assert len(weave.versions(latest_obj)) == 1


@pytest.mark.skip("was used for table caching which is disabled")
def test_none_not_cached():
    os.environ["WEAVE_CACHE_MODE"] = "minimal"

    input_vals = [1, None]
    expected_result = [10001, None]

    l = weave.save(input_vals)
    mapped = l.map(lambda x: expensive_op(x))
    res = weave.use(mapped)
    assert res == expected_result

    latest_obj = weave.use(
        weave.legacy.weave.ops.get("local-artifact:///run-op-expensive_op:latest/obj")
    )
    assert len(latest_obj) == 1  # not 2! None not cached!
    assert len(weave.versions(latest_obj)) == 1



from typing import Callable

LOG_ONCE_MESSAGE_SUFFIX = " (subsequent messages of this type will be suppressed)"
logged_messages = []


def log_once(log_method: Callable[[str], None], message: str) -> None:
    """Logs a message once, suppressing subsequent messages of the same type. This
    is useful for notifying the user about errors without spamming the logs.

    This is mostly useful for cases where the same error message might occur many times.
    For example, if an op fails to save, it is likely going to happen every time that op is
    called. Or, if we have an error in our patched iterator, then it likely happens every time
    we iterate over the result. This allows use to inform the user about the error without
    clogging up their logs.

    Args:
        log_method: The method to use to log the message. This should accept a string argument.
        message: The message to log.

    Example:
    ```python
    log_once(logger.error, "Failed to save op")
    ```
    """
    if message not in logged_messages:
        log_method(message + LOG_ONCE_MESSAGE_SUFFIX)
        logged_messages.append(message)



import typing

from weave.legacy.weave import op_def_type


class MyTypedDict(typing.TypedDict):
    a: int
    b: str
    c: typing.Union[int, str]


def test_generate_referenced_type_code():
    code = op_def_type.generate_referenced_type_code(list[MyTypedDict])
    assert (
        code
        == """class MyTypedDict(typing.TypedDict):
    a: int
    b: str
    c: typing.Union[int, str]
"""
    )



# WeaveJS ops used for testing. These are not used in production.

from weave.legacy.weave import graph, weave_internal
from weave.legacy.weave import weave_types as types
from weave.legacy.weave._dict_utils import typeddict_pick_output_type
from weave.legacy.weave.language_features.tagging import tagged_value_type
from weave.legacy.weave.language_features.tagging.tagging_op_logic import (
    op_get_tag_type_resolver,
    op_make_type_tagged_resolver,
)


def ensure_node(v):
    if isinstance(v, graph.Node):
        return v
    return weave_internal.const(v)


def weavejs_pick(obj: graph.Node, key: str):
    raw_output_type = typeddict_pick_output_type(
        {"self": obj.type, "key": types.Const(types.String(), key)}
    )
    output_type = op_make_type_tagged_resolver(
        raw_output_type, op_get_tag_type_resolver(obj.type)
    )
    return weave_internal.make_output_node(
        output_type,
        "pick",
        {"obj": obj, "key": graph.ConstNode(types.String(), key)},
    )


def count(arr):
    arr_node = ensure_node(arr)
    return weave_internal.make_output_node(
        types.Number(),
        "count",
        {
            "arr": arr_node,
        },
    )


def index(arr, index):
    arr_node = ensure_node(arr)
    index_node = ensure_node(index)
    output_type = op_make_type_tagged_resolver(
        arr_node.type.object_type, op_get_tag_type_resolver(arr_node.type)
    )
    return weave_internal.make_output_node(
        output_type,
        "index",
        {
            "arr": arr_node,
            "index": index_node,
        },
    )


def filter(arr, filterFn):
    arr_node = ensure_node(arr)
    return weave_internal.make_output_node(
        arr_node.type,
        "filter",
        {
            "arr": arr_node,
            "filterFn": ensure_node(filterFn),
        },
    )


def map(arr, mapFn):
    arr_node = ensure_node(arr)
    return weave_internal.make_output_node(
        arr_node.type,
        "map",
        {
            "arr": arr_node,
            "mapFn": ensure_node(mapFn),
        },
    )


def sort(arr, compFn, columnDirs):
    arr_node = ensure_node(arr)
    return weave_internal.make_output_node(
        arr_node.type,
        "sort",
        {
            "arr": arr_node,
            "compFn": ensure_node(compFn),
            "columnDirs": ensure_node(columnDirs),
        },
    )


def groupby(arr, groupByFn):
    arr_node = ensure_node(arr)
    groupByFn_node = ensure_node(groupByFn)
    return weave_internal.make_output_node(
        types.List(
            tagged_value_type.TaggedValueType(
                types.TypedDict({"groupKey": groupByFn_node.type.output_type}),
                types.List(arr_node.type.object_type),
            )
        ),
        "groupby",
        {"arr": arr_node, "groupByFn": groupByFn_node},
    )


def offset(arr, offset):
    arr_node = ensure_node(arr)
    return weave_internal.make_output_node(
        arr_node.type,
        "offset",
        {
            "arr": arr_node,
            "offset": ensure_node(offset),
        },
    )


def limit(arr, limit):
    arr_node = ensure_node(arr)
    return weave_internal.make_output_node(
        arr_node.type,
        "limit",
        {
            "arr": arr_node,
            "limit": ensure_node(limit),
        },
    )


def file_type(file):
    file_node = ensure_node(file)
    output_type = op_make_type_tagged_resolver(
        types.TypeType(), op_get_tag_type_resolver(file_node.type)
    )
    return weave_internal.make_output_node(
        output_type,
        "file-type",
        {
            "file": file_node,
        },
    )



import importlib
import typing

import weave
from weave.trace.op_extensions.accumulator import add_accumulator
from weave.trace.patcher import MultiPatcher, SymbolPatcher

if typing.TYPE_CHECKING:
    from litellm.utils import ModelResponse


# This accumulator is nearly identical to the mistral accumulator, just with different types.
def litellm_accumulator(
    acc: typing.Optional["ModelResponse"],
    value: "ModelResponse",
) -> "ModelResponse":
    # This import should be safe at this point
    from litellm.utils import Choices, Message, ModelResponse, Usage

    if acc is None:
        acc = ModelResponse(
            id=value.id,
            object=value.object,
            created=value.created,
            model=value.model,
            choices=[],
            usage=Usage(prompt_tokens=0, total_tokens=0, completion_tokens=None),
        )

    # Merge in the usage info
    if "usage" in value.model_fields_set and value.usage is not None:
        acc.usage.prompt_tokens += value.usage.prompt_tokens
        acc.usage.total_tokens += value.usage.total_tokens
        if acc.usage.completion_tokens is None:
            acc.usage.completion_tokens = value.usage.completion_tokens
        else:
            acc.usage.completion_tokens += value.usage.completion_tokens

    # Loop through the choices and add their deltas
    for delta_choice in value.choices:
        while delta_choice.index >= len(acc.choices):
            acc.choices.append(
                Choices(
                    index=len(acc.choices),
                    message=Message(role="", content=""),
                    finish_reason=None,
                )
            )
        acc.choices[delta_choice.index].message.role = (
            delta_choice.delta.role or acc.choices[delta_choice.index].message.role
        )
        acc.choices[delta_choice.index].message.content += (
            delta_choice.delta.content or ""
        )
        if delta_choice.delta.tool_calls:
            if acc.choices[delta_choice.index].message.tool_calls is None:
                acc.choices[delta_choice.index].message.tool_calls = []
            acc.choices[
                delta_choice.index
            ].message.tool_calls += delta_choice.delta.tool_calls
        acc.choices[delta_choice.index].finish_reason = (
            delta_choice.finish_reason or acc.choices[delta_choice.index].finish_reason
        )

    return acc


# LiteLLM does so odd stuff with pydantic objects which result in our auto
# serialization not working correctly. Here we just blindly dump to a dict instead.
def litellm_on_finish_post_processor(value: typing.Any) -> typing.Any:
    import pydantic

    value_to_finish = value
    if isinstance(value, pydantic.BaseModel):
        value_to_finish = value.model_dump()

    return value_to_finish


# Unlike other integrations, streaming is based on input flag, not
def should_use_accumulator(inputs: typing.Dict) -> bool:
    return isinstance(inputs, dict) and bool(inputs.get("stream"))


def make_wrapper(name: str) -> typing.Callable:
    def litellm_wrapper(fn: typing.Callable) -> typing.Callable:
        op = weave.op()(fn)
        op.name = name  # type: ignore
        return add_accumulator(
            op,  # type: ignore
            make_accumulator=lambda inputs: litellm_accumulator,
            should_accumulate=should_use_accumulator,
            on_finish_post_processor=litellm_on_finish_post_processor,
        )

    return litellm_wrapper


litellm_patcher = MultiPatcher(
    [
        SymbolPatcher(
            lambda: importlib.import_module("litellm"),
            "completion",
            make_wrapper("litellm.completion"),
        ),
        SymbolPatcher(
            lambda: importlib.import_module("litellm"),
            "acompletion",
            make_wrapper("litellm.acompletion"),
        ),
    ]
)



import weave
from weave.legacy.weave import box, graph
from weave.legacy.weave import context_state as _context_state
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.language_features.tagging import tag_store
from weave.legacy.weave.language_features.tagging.tagged_value_type import TaggedValueType

tag_adders = 0


def op_add_tag(obj_node: graph.Node, tags: dict[str, str]):  # type: ignore[no-untyped-def]
    global tag_adders
    tag_adders += 1
    name = f"_custom_tagger_{tag_adders}"

    _loading_builtins_token = _context_state.set_loading_built_ins()

    @weave.op(
        name=name,
        input_type={"obj": types.Any()},
        output_type=lambda input_types: TaggedValueType(
            types.TypedDict({f"_ct_{k}": types.String() for k in tags.keys()}),
            input_types["obj"],
        ),
    )
    def custom_tagger(obj):  # type: ignore[no-untyped-def]
        return tag_store.add_tags(
            box.box(obj), {f"_ct_{k}": v for k, v in tags.items()}
        )

    _context_state.clear_loading_built_ins(_loading_builtins_token)

    return custom_tagger(obj_node)


def make_get_tag(tag_name: str):  # type: ignore[no-untyped-def]
    from weave.legacy.weave.language_features.tagging import make_tag_getter_op

    return make_tag_getter_op.make_tag_getter_op(f"_ct_{tag_name}", types.String())
